#include "rpi-gpio.h"

#define RESET_TIMEOUT 10000
#define LOADER_ADDR 0x40D000

#define GP_LED (1 << 15)
#define GP_SYNC (1 << 8)
#define GP_MOSI (1 << 9)
#define GP_MISO (1 << 10)
#define GP_SCLK (1 << 11)

typedef unsigned char u8;
typedef unsigned short u16;

extern void _reset();

// External data included via xeno_qcode.S
extern u8 qcode[];
extern u8 *qcode_end;
extern u8 upload[];
extern u8 *upload_end;
extern u8 credits[];
extern u8 *credits_end;

unsigned delay_value = 300;

void ldelay(volatile unsigned i)
{
	if (i)
		while (i--);
}

void delay()
{
	volatile unsigned time;
	time = delay_value;
	if (delay_value)
		while (time--);
}

/*
 * SPI stuff
 * These functions bitbang the GPIO pins
 * Least significant bit first
 */

// Exchange a bit and tick the clock
unsigned io(unsigned i)
{
	unsigned res = 0;

	if (i)
		gpio[GPIO_GPSET0] = GP_MOSI;
	else
		gpio[GPIO_GPCLR0] = GP_MOSI;

	gpio[GPIO_GPCLR0] = GP_SCLK;
	delay();

	res = gpio[GPIO_GPLEV0] & GP_MISO;

	gpio[GPIO_GPSET0] = GP_SCLK;
	delay();

	return !!res;
}

// Exchange a byte
u8 io8(u8 out)
{
	u8 in = 0;
	unsigned i;

	for (i = 0; i < 8; ++i)
	{
		in |= io(out & (1 << i)) << i;
	}

	return in;
}

// Send a byte with firmware specific waits
void send8(u8 out)
{
	unsigned i = 0;

	while (gpio[GPIO_GPLEV0] & GP_SYNC)
	{
		if (i++ > RESET_TIMEOUT)
			_reset();
	}

	for (i = 0; i < 8; ++i)
	{
		if (gpio[GPIO_GPLEV0] & GP_SYNC)
			_reset();
		io(out & (1 << i));
	}
}

// Receive a byte with firmware specific waits
u8 recv8()
{
	unsigned i = 0;
	u8 in = 0;

	while (!(gpio[GPIO_GPLEV0] & GP_SYNC))
	{
		if (i++ > RESET_TIMEOUT)
			_reset();
	}

	for (i = 0; i < 8; ++i)
	{
		if (!(gpio[GPIO_GPLEV0] & GP_SYNC))
			_reset();
		in |= io(0) << i;
	}

	return in;
}

/*
 * Helper functions for firmware commands
 */

unsigned read_mem(u8 *dst, unsigned addr, u8 len)
{
	unsigned err;

	send8(0xFF);
	send8(0);
	send8(addr >> 8);
	send8(addr);
	send8(0);
	send8(addr >> 16);
	send8(0);
	send8(0);

	send8(len);
	send8(0);

	err = recv8();
	err |= recv8() << 8;

	if (err)
		return err;

	while (len--)
		*dst++ = recv8();

	return 0;
}

void write_word_norecv(unsigned addr, u16 data)
{
	send8(0xFE);
	send8(0);
	send8(addr >> 8);
	send8(addr);

	send8(data);
	send8(addr >> 16);
	send8(data >> 8);
	send8(0);

	send8(2);
	send8(0);
}

// The firmware sends back 4 bytes when writing a word, discard them silently
void write_word(unsigned addr, u16 data)
{
	write_word_norecv(addr, data);

	recv8();
	recv8();
	recv8();
	recv8();
}

void write_block(unsigned addr, u8 *source, int len)
{
	while (len >= 1)
	{
		// The firmware writes words in little endian order
		write_word(addr, source[0] | (source[1] << 8));
		addr += 2;
		len -= 2;
		source += 2;
	}
}

void kernel_main(unsigned r0, unsigned r1, unsigned atags)
{
	const unsigned qcodesize = ((((unsigned) &qcode_end) - ((unsigned) &qcode)) & 0xFFFE) + 2;
	const unsigned uploadsize = ((((unsigned) &upload_end) - ((unsigned) &upload)) & 0xFFFE) + 2;
	const unsigned creditssize = ((((unsigned) &credits_end) - ((unsigned) &credits)) & 0xFFFE) + 2;

	u16 last_recv = 0, wUploadSize, wTest = 0;
	unsigned i;
	u8 r, e, n, *pUpload;

	// Bootstrap stage 1
	static u8 pLoaderCode[] =
	{
		0x80, 0x00,					// 8000			MOV $00,D0
		0xC4, 0xDA, 0xFC,			// C4DAFC		MOVB D0,($FCDA)		# disable breakpoints

		0xF4,0x74,0x74,0x0a,0x08,	// F47474A708	MOV $080a74,a0		# restore original
		0xF7,0x20,0x4C,0x80,		// F7204C80		MOV a0,($804c)		# inthandler
		0xF4,0x74,					// F47400D040	MOV QCODEIMGBASE,a0	# jump to drivecode init
		(LOADER_ADDR & 0xFF),
		(LOADER_ADDR >> 8 & 0xFF),
		(LOADER_ADDR >> 16 & 0xFF),
		0xF0,0x00					// F000			JMP (a0)
	};

	// Reset GPIO ports
	gpio[GPIO_GPFSEL0] = 0;
	gpio[GPIO_GPFSEL1] = 0;

	// Initialize the on-board ACT LED
	gpio[GPIO_GPFSEL1] |= (1 << 18);
	gpio[GPIO_GPCLR1] = GP_LED;

	// Set up GPIO to communicate with the DVD drive
	// GPIO	Role			Mode
	// 8	PB1 (SYNC)		input
	// 9	PB4 (MOSI)		output
	// 10	PB3 (MISO)		input
	// 11	PB5 (SCLK)		output
	gpio[GPIO_GPFSEL0] |= (1 << 27);
	gpio[GPIO_GPFSEL1] |= (1 << 3);

	// Initialize the output pins
	gpio[GPIO_GPCLR0] = GP_SCLK | GP_MOSI;

	// Wait for the firmware to be ready
	do
	{
		last_recv >>= 1;
		last_recv |= io(1) ? 0x8000 : 0;
	}
	while (last_recv != 0xEEEE);

	pUpload = qcode;
	wUploadSize = qcodesize;

	// Check whether the "credits binary" was requested
	read_mem((u8 *) &wTest, 0x40D100, 2);
	if (wTest == 0x4444)
	{
		pUpload = credits;
		wUploadSize = creditssize;
	}

	// Bootstrap stage 2
	write_block(LOADER_ADDR, (u8 *) upload, uploadsize);

	// Bootstrap stage 1
	write_block(0x8674, pLoaderCode, sizeof(pLoaderCode));

	// Redirect a jump toward the bootstrap
	write_word_norecv(0x804D, 0x0086);

	// Effect unknown
	// Probably initializing the transfer with the bootstrap
	io8(0);

	// I'm not actually modifying the delay here
	delay_value = 300;

	ldelay(300);
	io8(0);

	ldelay(300);
	io8((wUploadSize >> 9) & 0xFF);
	ldelay(300);
	io8((wUploadSize >> 1) & 0xFF);
	ldelay(300);
	io8(0);

	// Turn on the LED before starting the actual transfer
	gpio[GPIO_GPSET1] = GP_LED;

	// First byte that comes back
	e = (wUploadSize >> 1) & 0xFF;

	// And upload the drive code
	for (i = 0; i < wUploadSize; ++i)
	{
		ldelay(300);
		r = io8(n = pUpload[i]);

		if (r != e)
			_reset();

		e = n;
	}

	// Probably ending the transfer
	ldelay(300);

	io8(0x21);

	io8(0);
	io8(0);

	// Reset the chip here, it'll wait for the next drive reset to patch again
	_reset();
}

